#ifndef	V_IMATH_H
#define	V_IMATH_H

/*      ================================================================
 *      Vector operations
 *      ================================================================
 */

/*		Assignment	*/
#define	vec2Set(vec,x,y)\
		(vec)[0]=(x);\
		(vec)[1]=(y);

#define	vec3Set(vec,x,y,z)\
		(vec)[0]=(x);\
		(vec)[1]=(y);\
		(vec)[2]=(z);

#define	vec4Set(vec,x,y,z,w)\
		(vec)[0]=x;\
		(vec)[1]=y;\
		(vec)[2]=z;\
		(vec)[3]=w;

/*		Setting to zero	*/
#define	vec2Zero(vec)\
		vec2Set(vec,0,0)

#define	vec3Zero(vec)\
		vec3Set(vec,0,0,0);

#define	vec4Zero(vec)\
		vec4Set(vec,0,0,0,0);

/*		Copying	*/
#define	vec2Copy(dst,src)\
		(dst)[0]=(src)[0];\
		(dst)[1]=(src)[1];

#define	vec3Copy(dst,src)\
		(dst)[0]=(src)[0];\
		(dst)[1]=(src)[1];\
		(dst)[2]=(src)[2];

#define	vec4Copy(dst,src)\
		(dst)[0]=(src)[0];\
		(dst)[1]=(src)[1];\
		(dst)[2]=(src)[2];\
		(dst)[3]=(src)[3];

/*		Addition	*/
#define	vec2Add(dst,v1,v2)\
		(dst)[0]=(v1)[0]+(v2)[0];\
		(dst)[1]=(v1)[1]+(v2)[1];

#define	vec3Add(dst,v1,v2)\
		(dst)[0]=(v1)[0]+(v2)[0];\
		(dst)[1]=(v1)[1]+(v2)[1];\
		(dst)[2]=(v1)[2]+(v2)[2];

#define	vec4Add(dst,v1,v2)\
		(dst)[0]=(v1)[0]+(v2)[0];\
		(dst)[1]=(v1)[1]+(v2)[1];\
		(dst)[2]=(v1)[2]+(v2)[2];\
		(dst)[3]=(v1)[3]+(v2)[3];

/*		Subtraction	*/
#define	vec2Sub(dst,v1,v2)\
		(dst)[0]=(v1)[0]-(v2)[0];\
		(dst)[1]=(v1)[1]-(v2)[1];

#define	vec3Sub(dst,v1,v2)\
		(dst)[0]=(v1)[0]-(v2)[0];\
		(dst)[1]=(v1)[1]-(v2)[1];\
		(dst)[2]=(v1)[2]-(v2)[2];

#define	vec4Sub(dst,v1,v2)\
		(dst)[0]=(v1)[0]-(v2)[0];\
		(dst)[1]=(v1)[1]-(v2)[1];\
		(dst)[2]=(v1)[2]-(v2)[2];\
		(dst)[3]=(v1)[3]-(v2)[3];

/*		Multiplication by scalar	*/
#define	vec2Scale(dst,vec,ml)\
		(dst)[0]=(vec)[0]*(ml);\
		(dst)[1]=(vec)[1]*(ml);

#define	vec3Scale(dst,vec,ml)\
		(dst)[0]=(vec)[0]*(ml);\
		(dst)[1]=(vec)[1]*(ml);\
		(dst)[2]=(vec)[2]*(ml);

#define	vec4Scale(dst,vec,ml)\
		(dst)[0]=(vec)[0]*(ml);\
		(dst)[1]=(vec)[1]*(ml);\
		(dst)[2]=(vec)[2]*(ml);\
		(dst)[3]=(vec)[3]*(ml);

/*		Multiplication by -1 not in-place	*/
#define	vec2Negate(dst,src)\
		(dst)[0]=-(src)[0];\
		(dst)[1]=-(src)[1];

#define	vec3Negate(dst,src)\
		(dst)[0]=-(src)[0];\
		(dst)[1]=-(src)[1];\
		(dst)[2]=-(src)[2];

#define	vec4Negate(dst,src)\
		(dst)[0]=-(src)[0];\
		(dst)[1]=-(src)[1];\
		(dst)[2]=-(src)[2];\
		(dst)[3]=-(src)[3];

/*    	Dot products */
#define	vec2Dot(v1,v2)\
		((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1])

#define	vec3Dot(v1,v2)\
		((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2])

#define	vec4Dot(v1,v2)\
		((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2]+(v1)[3]*(v2)[3])

/*		Vector length */
#define	vec2Abs(vec)\
		sqrt(vec2Dot(vec,vec))

#define	vec3Abs(vec)\
		sqrt(vec3Dot(vec,vec))

#define	vec4Abs(vec)\
		sqrt(vec4Dot(vec,vec))

/*    These macros make unit vectors (not in place) */
#define	vec2Unit(dst,src)\
		{\
			vec_t ra;\
			ra=(vec_t)1.0/vec2Abs(src);\
			vec2Scale(dst,src,ra);\
		}

#define	vec3Unit(dst,src)\
		{\
			vec_t ra;\
			ra=(vec_t)1.0/vec3Abs(src);\
			vec3Scale(dst,src,ra);\
		}

/*		This macro makes homogenous vector */
#define	vec4Homogenize(dst,src)\
		{\
			vec_t rw;\
			rw = (vec_t)1.0 / (src)[3];\
			vec3Scale((vec3_t)(dst),(vec3_t)(src),rw);\
			(dst)[3] = (vec_t)1.0;\
		}

/*		Cross product in 2-D space (oriented parallelogram area)		*/
#define	vec2Cross(v1,v2)\
		((v1)[0] * (v2)[1] - (v1)[1] * (v2)[0])

/*		Cross product in 3-D space (oriented parallelepiped volume)		*/
#define	vec3Cross(dst,v1,v2)\
		(dst)[0]=(v1)[1]*(v2)[2]-(v1)[2]*(v2)[1];\
		(dst)[1]=(v1)[2]*(v2)[0]-(v1)[0]*(v2)[2];\
		(dst)[2]=(v1)[0]*(v2)[1]-(v1)[1]*(v2)[0];

/*		These macros return indices of minimal/maximal components		*/
#define	vec2MinAbsInd(vec)\
		((v_abs((vec)[0]) < v_abs((vec)[1])) ? 0 : 1)

#define	vec2MaxAbsInd(vec)\
		((v_abs((vec)[0]) > v_abs((vec)[1])) ? 0 : 1)

/*		This returns true if argument is almost zero						*/
#define	IsZero(v)\
		((v_abs(v) < FLOAT_ZERO_EPSILON) ? true : false)

/*		These return true if length of vector is almost zero			*/
#define vec2IsZero(vec)\
		(vec2Dot(vec, vec) < FLOAT_ZERO_EPSILON_SQUARE) ? true : false)

#define vec3IsZero(vec)\
		(vec3Dot(vec, vec) < FLOAT_ZERO_EPSILON_SQUARE) ? true : false)

#define vec4IsZero(vec)\
		(vec4Dot(vec, vec) < FLOAT_ZERO_EPSILON_SQUARE) ? true : false)

/*		These return true if vectors are almost equal					*/
#define	IsEqualEpsilon(v1,v2,epsilon)\
		(v_abs(v1-v2) <= epsilon)

#define	vec2IsEqual(v1,v2)\
		(IsZero((v1)[0]-(v2)[0]) && IsZero((v1)[1]-(v2)[1]))

#define	vec3IsEqual(v1,v2)\
		(IsZero((v1)[0]-(v2)[0]) && IsZero((v1)[1]-(v2)[1] && \
		 IsZero((v1)[2]-(v2)[2]))

#define	vec4IsEqual(v1,v2)\
		(IsZero((v1)[0]-(v2)[0]) && IsZero((v1)[1]-(v2)[1] && \
		 IsZero((v1)[2]-(v2)[2]) && IsZero((v1)[2]-(v2)[2]))

/*		These make 2/3/4 products of components of args					*/
#define	vec2VecScale(dst,v1,v2)\
		(dst)[0] = (v1)[0] * (v2)[0];\
		(dst)[1] = (v1)[1] * (v2)[1];

#define	vec3VecScale(dst,v1,v2)\
		(dst)[0] = (v1)[0] * (v2)[0];\
		(dst)[1] = (v1)[1] * (v2)[1];\
		(dst)[2] = (v1)[2] * (v2)[2];

#define	vec4VecScale(dst,v1,v2)\
		(dst)[0] = (v1)[0] * (v2)[0];\
		(dst)[1] = (v1)[1] * (v2)[1];\
		(dst)[2] = (v1)[2] * (v2)[2];\
		(dst)[3] = (v1)[3] * (v2)[3];

/*		These make 2/3/4 quotiens of components of args. Useful when one
 *	need to construct scale factors by given initial and final sizes.
 */
#define	vec2ZoomIF(zoom,initial,final)\
		(zoom)[0] = (final)[0] / (initial)[0];\
		(zoom)[1] = (final)[1] / (initial)[1];

#define	vec3ZoomIF(zoom,initial,final)\
		(zoom)[0] = (final)[0] / (initial)[0];\
		(zoom)[1] = (final)[1] / (initial)[1];\
		(zoom)[2] = (final)[2] / (initial)[2];

#define	vec4ZoomIF(zoom,initial,final)\
		(zoom)[0] = (final)[0] / (initial)[0];\
		(zoom)[1] = (final)[1] / (initial)[1];\
		(zoom)[2] = (final)[2] / (initial)[2];\
		(zoom)[3] = (final)[3] / (initial)[3];

/*		Conversion from degrees to radians								*/
#define	vec2ToRadians(dst,src)\
		vec2Scale(dst,src,ANGLE_TO_RAD);

#define	vec3ToRadians(dst,src)\
		vec3Scale(dst,src,ANGLE_TO_RAD);

#define	vec4ToRadians(dst,src)\
		vec4Scale(dst,src,ANGLE_TO_RAD);

/*		Conversion from radians to degrees								*/
#define	vec2ToDegrees(dst,src)\
		vec2Scale(dst,src,RAD_TO_ANGLE);

#define	vec3ToDegrees(dst,src)\
		vec3Scale(dst,src,RAD_TO_ANGLE);

#define	vec4ToDegrees(dst,src)\
		vec4Scale(dst,src,RAD_TO_ANGLE);

/*	================================================================
 *	Matrix operations
 *	================================================================
 */

/*		Setting to zero	*/
#define	mat2Zero(mat)\
		memZero32(mat,sizeof(vec_t));

#define	mat3Zero(mat)\
		memZero32(mat,9 * (sizeof(vec_t)>>2));

#define	mat4Zero(mat)\
		memZero32(mat,sizeof(vec_t)<<2);

#define	mat3x4Zero(mat)\
		memZero32(mat,3*sizeof(vec_t));

/*		Copying */
#define	mat2Copy(dst,src)\
		memCopy32(dst,src,sizeof(vec_t));

#define	mat3Copy(dst,src)\
		memCopy32((dst),src,9*(sizeof(vec_t)>>2));

#define	mat4Copy(dst,src)\
		memCopy32(dst,src,sizeof(vec_t)<<2);

#define	mat3x4Copy(dst,src)\
		memCopy32(dst,src,3*sizeof(vec_t));

/*		Setting to identity state */
#define	mat2Unit(mat)\
		(mat)[1][0]=(mat)[0][1]=0.0;\
		(mat)[0][0]=(mat)[1][1]=1.0;

#define	mat3Unit(mat)\
		mat3Zero(mat);\
		(mat)[0][0]=(mat)[1][1]=(mat)[2][2]=1.0;

#define	mat4Unit(mat)\
		mat4Zero(mat);\
		(mat)[0][0]=(mat)[1][1]=(mat)[2][2]=(mat)[3][3]=1.0;

#define	mat3x4Unit(mat)\
		mat3x4Zero(mat);\
		(mat)[0][0]=(mat)[1][1]=(mat)[2][2]=1.0;


/*		This copies	33-matrix into	LU-corner of 44/3x4-matrix	*/
#define	mat4SetLU(dst,LU)\
		(dst)[0][0]=(LU)[0][0];(dst)[0][1]=(LU)[0][1];\
			(dst)[0][2]=(LU)[0][2];\
		(dst)[1][0]=(LU)[1][0];(dst)[1][1]=(LU)[1][1];\
			(dst)[1][2]=(LU)[1][2];\
		(dst)[2][0]=(LU)[2][0];(dst)[2][1]=(LU)[2][1];\
			(dst)[2][2]=(LU)[2][2];

/*		This copies	LU-corner of 44/34-matrix	into 33-matrix	*/
#define	mat3GetLU(LU,src)\
		(LU)[0][0]=(src)[0][0];(LU)[0][1]=(src)[0][1];\
			(LU)[0][2]=(src)[0][2];\
		(LU)[1][0]=(src)[1][0];(LU)[1][1]=(src)[1][1];\
			(LU)[1][2]=(src)[1][2];\
		(LU)[2][0]=(src)[2][0];(LU)[2][1]=(src)[2][1];\
			(LU)[2][2]=(src)[2][2];

/*		This inserts three 3-vectors into rows of 3x3-matrix	*/
#define	mat3SetRows(mat,v1,v2,v3)\
		(mat)[0][0]=(v1)[0]; (mat)[0][1]=(v1)[1]; (mat)[0][2]=(v1)[2];\
		(mat)[1][0]=(v2)[0]; (mat)[1][1]=(v2)[1]; (mat)[1][2]=(v2)[2];\
		(mat)[2][0]=(v3)[0]; (mat)[2][1]=(v3)[1]; (mat)[2][2]=(v3)[2];

/*		Inserts	three 3-vectors	into cols of 3x3-matrix	*/
#define	mat3SetCols(mat,v1,v2,v3)\
		(mat)[0][0]=(v1)[0]; (mat)[0][1]=(v2)[0]; (mat)[0][2]=(v3)[0];\
		(mat)[1][0]=(v1)[1]; (mat)[1][1]=(v2)[1]; (mat)[1][2]=(v3)[1];\
		(mat)[2][0]=(v1)[2]; (mat)[2][1]=(v2)[2]; (mat)[2][2]=(v3)[2];

/*		Inserts	three 3-vectors	into rows of 3x4-matrix and sets last
		column of matrix to zero */
#define	mat3x4SetRows(mat,v1,v2,v3)\
		(mat)[0][0]=(v1)[0]; (mat)[0][1]=(v1)[1]; (mat)[0][2]=(v1)[2];\
			(mat)[0][3]=0;\
		(mat)[1][0]=(v2)[0]; (mat)[1][1]=(v2)[1]; (mat)[1][2]=(v2)[2];\
			(mat)[1][3]=0;\
		(mat)[2][0]=(v3)[0]; (mat)[2][1]=(v3)[1]; (mat)[2][2]=(v3)[2];\
			(mat)[2][3]=0;

/*		Inserts	four 3-vectors into	cols of	3x4-matrix	*/
#define mat3x4SetCols(mat,v1,v2,v3,v4)\
		(mat)[0][0]=(v1)[0]; (mat)[0][1]=(v2)[0]; (mat)[0][2]=(v3)[0];\
			(mat)[0][3]=(v4)[0];\
		(mat)[1][0]=(v1)[1]; (mat)[1][1]=(v2)[1]; (mat)[1][2]=(v3)[1];\
			(mat)[1][3]=(v4)[1];\
		(mat)[2][0]=(v1)[2]; (mat)[2][1]=(v2)[2]; (mat)[2][2]=(v3)[2];\
			(mat)[2][3]=(v4)[2];

/*      Transposition of matrix2_t not-in-place */
#define	mat2Transpose(dst,src)\
		dst[0][0] = (src)[0][0];\
		dst[0][1] = (src)[1][0];\
		dst[1][0] = (src)[0][1];\
		dst[1][1] = (src)[1][1];

/*		Transposition of matrix2_t in-place */
#define	mat2TransposeIP(mat)\
		{\
			vec_t temp;\
			temp=(mat)[0][1];(mat)[0][1]=(mat)[1][0];(mat)[1][0]=temp;\
		}

/*		Matrix by vector multiplication */
#define	mat2VecMul(dst,mat,v)\
		(dst)[0]=(mat)[0][0]*(v)[0]+(mat)[0][1]*(v)[1];\
		(dst)[1]=(mat)[1][0]*(v)[0]+(mat)[1][1]*(v)[1];

#define	mat3VecMul(dst,mat,v)\
		(dst)[0]=(mat)[0][0]*(v)[0]+(mat)[0][1]*(v)[1]+mat[0][2]*(v)[2];\
		(dst)[1]=(mat)[1][0]*(v)[0]+(mat)[1][1]*(v)[1]+mat[1][2]*(v)[2];\
		(dst)[2]=(mat)[2][0]*(v)[0]+(mat)[2][1]*(v)[1]+mat[2][2]*(v)[2];

#define	mat4VecMul(dst,mat,v)\
		(dst)[0]=(mat)[0][0]*(v)[0]+(mat)[0][1]*(v)[1]+(mat)[0][2]*(v)[2]+\
			(mat)[0][3]*(v)[3];\
		(dst)[1]=(mat)[1][0]*(v)[0]+(mat)[1][1]*(v)[1]+(mat)[1][2]*(v)[2]+\
			(mat)[1][3]*(v)[3];\
		(dst)[2]=(mat)[2][0]*(v)[0]+(mat)[2][1]*(v)[1]+(mat)[2][2]*(v)[2]+\
			(mat)[2][3]*(v)[3];\
		(dst)[3]=(mat)[3][0]*(v)[0]+(mat)[3][1]*(v)[1]+(mat)[3][2]*(v)[2]+\
			(mat)[3][3]*(v)[3];

/*		Matrix 3x4 by 3 - vector "multiplication". Actualy, there is a
 *	matrix3_t by vec3_t multiplication (for R+D) combined with translation.
 */

#define	mat3x4VecMul(dst,mat,v)\
		(dst)[0]=(mat)[0][0]*(v)[0]+(mat)[0][1]*(v)[1]+(mat)[0][2]*(v)[2]+\
			(mat)[0][3];\
		(dst)[1]=(mat)[1][0]*(v)[0]+(mat)[1][1]*(v)[1]+(mat)[1][2]*(v)[2]+\
			(mat)[1][3];\
		(dst)[2]=(mat)[2][0]*(v)[0]+(mat)[2][1]*(v)[1]+(mat)[2][2]*(v)[2]+\
			(mat)[2][3];

/*		Multiplication of two matrices of sizes 2x2 */
#define	mat2MatMul(dst,m1,m2)\
		(dst)[0][0]=(m1)[0][0]*(m2)[0][0]+(m1)[0][1]*(m2)[1][0];\
		(dst)[0][1]=(m1)[0][0]*(m2)[0][1]+(m1)[0][1]*(m2)[1][1];\
		(dst)[1][0]=(m1)[1][0]*(m2)[0][0]+(m1)[1][1]*(m2)[1][0];\
		(dst)[1][1]=(m1)[1][0]*(m2)[0][1]+(m1)[1][1]*(m2)[1][1];

/*		Concatenations of two transforms	*/
#define	mat3x4ConcatBefore(dst,transform)\
		{\
			matrix3x4_t tmp;\
			mat3x4ByMat(tmp,(transform),(dst));\
			mat3x4Copy((dst),tmp);\
		}

#define	mat3x4ConcatAfter(dst,transform)\
		{\
			matrix3x4_t tmp;\
			mat3x4ByMat3x4(tmp,(dst),(transform));\
			mat3x4Copy((dst),tmp);\
		}

/*	======================================================================
 *	Color operations
 *	======================================================================
 */

#define	rgbFromByte(rgb,brgb)\
		rgb[0]=brgb[0];\
		rgb[1]=brgb[1];\
		rgb[2]=brgb[2];\
		vec3Scale(rgb,rgb,1/(vec_t)255);

#define	rgbaFromByte(rgba,brgba)\
		rgba[0]=brgba[0];\
		rgba[1]=brgba[1];\
		rgba[2]=brgba[2];\
		rgba[3]=brgba[3];\
		vec4Scale(rgba,rgba,1/(vec_t)255);

#define	rgbaByteCopy(dest,src)\
		*((u32_t*)(dest))=*(u32_t*)(src);

#define	rgbByteFromFloat(brgb,rgb) \
		{ \
			vec3_t scaled; \
			vec3Scale(scaled,(rgb),255); \
			(brgb)[0]=CLAMP_FLOAT_TO_BYTE(scaled[0]); \
			(brgb)[1]=CLAMP_FLOAT_TO_BYTE(scaled[1]); \
			(brgb)[2]=CLAMP_FLOAT_TO_BYTE(scaled[2]); \
		}

#define	rgbaByteFromFloat(brgba,rgba) \
		{ \
			vec4_t scaled; \
			vec4Scale(scaled,rgba,255); \
			(brgb)[0]=CLAMP_FLOAT_TO_BYTE(scaled[0]); \
			(brgb)[1]=CLAMP_FLOAT_TO_BYTE(scaled[1]); \
			(brgb)[2]=CLAMP_FLOAT_TO_BYTE(scaled[2]); \
			(brgb)[3]=CLAMP_FLOAT_TO_BYTE(scaled[3]); \
		}

#endif	/* V_IMATH_H */
