#include <../source/TAL/objbase.h>
#include <../source/TAL/fonthead.h>
#include <../source/geom/stroke.h>
#include <../source/vfs/mbase.h>
#include <../source/vfs/fswrap.h>

/*
 * Support for Borland stroke fonts.
 * Last update: 03.12.2002
 */

#ifdef	CHECKED_BUILD
char	*fh_object="font";
#endif	/* CHECKED_BUILD */

typedef struct stroke_font_s
{
	fonthead_t	head;
    u32_t   num_codes;  /* Number of codes (NOT glyphs !) in font */
    vec_t   *widths;    /* Width table */
    u32_t   *code_tab;  /* Mapping from input ABC to glyph offset */
    vhandle_t glyphs;  /* Glyph definitions */
} stroke_font_t, *stroke_font_p;

V_LOCAL u32_t V_CALL
bgi_destroy_font(vhandle_t self)
{
	stroke_font_p font=(stroke_font_p)self;
	UNREF_EXIT(&font->head);
	/* No more references: delete glyphs */
	meshDeleteStrokes(font->glyphs);
	/* Delete object */
    memFreeObject(self);
	return 0;
}

V_LOCAL	u32_t
bgi_glyph_of_char(stroke_font_t *font,u32_t ch)
{
    if(ch>=font->num_codes)
	{
        if('?'<font->num_codes)
            ch='?';
        else
            ch=' ';
	}
    return font->code_tab[ch];
}

V_LOCAL void V_CALL
bgi_get_stroke_extents(fonthead_t *handle,u32_t ch,vec_t *x,vec_t *y)
{
    stroke_font_t *font=(stroke_font_t *)handle;
    u32_t glyph=bgi_glyph_of_char(font,ch);
    if(x)
    	*x=font->widths[glyph];
    if(y)
    	*y=font->head.height;
}

/*
 * Draws glyph with code 'ch'.
 * Assumes 'font' is correct stroke font handle;
 * other args will be checked for validity.
 * Can optional store char width in 'dx_return'.
 */
V_LOCAL	void	V_CALL
bgi_draw_char(fonthead_t *handle,u32_t ch,vec_t *dx_return)
{
    vec_t dx;
    u32_t glyph,i,num_strokes,j,num_coords;
    mesh_stroke_t *stroke;
    stroke_font_t *font=(stroke_font_t *)handle;
    glyph=bgi_glyph_of_char(font,ch);
    num_strokes=meshStrokesCount(font->glyphs,glyph);
    dx=font->widths[glyph];
    for(i=0;i<num_strokes;i++)
    {
        stroke=meshGlyphStroke(font->glyphs,glyph,i);
        num_coords=stroke->num_coords;
        glBegin(GL_LINE_STRIP);
        for(j=0;j<num_coords;j++)
            VGL_VERTEX3V(stroke->coords[j]);
        glEnd();
    }
    glTranslatef((GLfloat)dx,0,0);
    if(dx_return)
        *dx_return=dx;
}

V_LOCAL stroke_font_p
bgi_convert_strokes_to_font(
    vec_t height,   /* Height of capital */
    vec_t ascent,   /* Ascent (positive) */
    vec_t descent,  /* Descent (positive or negative) */
    u32_t num_codes,
    vhandle_t glyphs)
{
    u32_t extra;
    stroke_font_t *fnt;
    u32_t num_glyphs=meshGlyphCount(glyphs);
    extra=num_glyphs*sizeof(vec_t)+ /* For width table */
        num_codes*sizeof(u32_t);    /* For char mapping */

    ALLOC_OBJECT(fnt,stroke_font_t,extra,FH_MAGIC);
    fnt->head.base.destroy=bgi_destroy_font;

    fnt->head.height=height;
    fnt->head.ascent=ascent;
    fnt->head.descent=descent;
    fnt->head.draw_char=bgi_draw_char;
    fnt->head.get_char_extents=bgi_get_stroke_extents;
	fnt->head.line_width=(vec_t)2;
    fnt->head.line_smooth=true;
	fnt->head.sp.scale=(vec_t)1;
	fnt->head.sp.fov=(vec_t)(M_PI/10);
	fnt->head.sp.use_fov=false;
	fnt->head.sp.adjustment=fa_left_top;
	vec3Set(fnt->head.sp.custom_adjustment,0,0,0);

    fnt->num_codes=num_codes;
    meshFinishStrokes(glyphs);
    fnt->glyphs=glyphs;
    fnt->widths=(vec_t*)(fnt+1);
    fnt->code_tab=(u32_t*)(fnt->widths+num_glyphs);

    /* NOTE: Caller have to setup code table and widths */
    return fnt;
}

/*  =================================================================== */

/*
 *  Input:
 *  Byte 1          7   6   5   4   3   2   1   0     bit #
 *                 op1  <seven bit signed X coord>
 *  Byte 2          7   6   5   4   3   2   1   0     bit #
 *                 op2  <seven bit signed Y coord>
 *      op1=0  op2=0  End of character definition.
 *      op1=1  op2=0  Move the pointer to (x,y)
 *      op1=1  op2=1  Draw from current pointer to (x,y)
 *  Returns stroke opcode: 0 - end; 1 - move to; 2 - line to.
 */
V_LOCAL u32_t
bgi_decode_stroke(u16_t stroke,vec3_t coord)
{
	/*
	 * This was hard, very, very hard ...
	 * But now it's (probably ?) endian - independent ...
	 * Please, feel free to correct this ...
	 * I don't want to muck with this anymore ... --katarn--
	 */
	int op1,op2;
	op1 = (stroke & (1 << 7));
	op2 = (stroke & (1 << 15));
	stroke <<= 1;
	coord[0] = 0.5f * (float)(signed char)(stroke & 0xFE);
	coord[1] = 0.5f * (float)(signed char)((stroke & 0xFE00) >> 8);
	coord[2] = 0;
    return op2 ? 2 : (op1 ? 1 : 0);
}

/*
 * Reads glyph definition from BGI font.
 */
V_LOCAL vboolean
bgi_read_glyph(vhandle_t in,vhandle_t out,vec_t scale)
{
    u32_t opcode=0;
    vec3_t coord, origin, pt;
    while(!fileEof(in))
    {
        opcode=bgi_decode_stroke(fileGetLittleShort(in),coord);
        if(opcode==2)
        {
            /* LineTo: store coord */
            vec3Scale(pt,coord,scale);
            meshStrokeCoordv(out,pt);
        }
        else
        {
            /* End of MoveTo: begin new stroke */
            meshEndStroke(out);
            if(!opcode) /* End */
                break;
            vec2Copy(origin,coord); /* MoveTo */
            vec3Scale(pt,coord,scale);
            meshStrokeCoordv(out,pt);
        }
    }
    if(opcode)
    {
        /* Error: unexpected EOF; discard all coords and strokes */
        meshCancelStroke(out);
        meshCancelGlyph(out);
    }
    else
    {
        /* Create glyph */
        meshEndGlyph(out);
    }
    return opcode==0;
}

V_LOCAL u32_t bgi_koi2alt[256]=
{
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
    0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
    0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
    0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
    0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
    0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
    0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
    0xc4, 0xb3, 0xda, 0xbf, 0xc0, 0xd9, 0xc3, 0xb4,
    0xc2, 0xc1, 0xc5, 0xdf, 0xdc, 0xdb, 0xdd, 0xde,
    0xb0, 0xb1, 0xb2, 0xf4, 0xfe, 0xf9, 0xfb, 0xf7,
    0xf3, 0xf2, 0xff, 0xf5, 0xf8, 0xfd, 0xfa, 0xf6,
    0xcd, 0xba, 0xd5, 0xf1, 0xd6, 0xc9, 0xb8, 0xb7,
    0xbb, 0xd4, 0xd3, 0xc8, 0xbe, 0xbd, 0xbc, 0xc6,
    0xc7, 0xcc, 0xb5, 0xf0, 0xb6, 0xb9, 0xd1, 0xd2,
    0xcb, 0xcf, 0xd0, 0xca, 0xd8, 0xd7, 0xce, 0xfc,
    0xee, 0xa0, 0xa1, 0xe6, 0xa4, 0xa5, 0xe4, 0xa3,
    0xe5, 0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae,
    0xaf, 0xef, 0xe0, 0xe1, 0xe2, 0xe3, 0xa6, 0xa2,
    0xec, 0xeb, 0xa7, 0xe8, 0xed, 0xe9, 0xe7, 0xea,
    0x9e, 0x80, 0x81, 0x96, 0x84, 0x85, 0x94, 0x83,
    0x95, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e,
    0x8f, 0x9f, 0x90, 0x91, 0x92, 0x93, 0x86, 0x82,
    0x9c, 0x9b, 0x87, 0x98, 0x9d, 0x99, 0x97, 0x9a
};

V_LOCAL u32_t bgi_win2alt[256]=
{
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,
    0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
    0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
    0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
    0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
    0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,
    0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
    0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,
    0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
    0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
    0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
    0x3f, 0x3f, 0x27, 0x3f, 0x22, 0x3a, 0xc5, 0xd8,
    0x3f, 0x25, 0x3f, 0x3c, 0x3f, 0x3f, 0x3f, 0x3f,
    0x3f, 0x27, 0x27, 0x22, 0x22, 0x07, 0x2d, 0x2d,
    0x3f, 0x54, 0x3f, 0x3e, 0x3f, 0x3f, 0x3f, 0x3f,
    0xff, 0xf6, 0xf7, 0x3f, 0xfd, 0x3f, 0xb3, 0x15,
    0xf0, 0x63, 0xf2, 0x3c, 0xbf, 0x2d, 0x52, 0xf4,
    0xf8, 0x2b, 0x49, 0x69, 0x3f, 0xe7, 0x14, 0xfa,
    0xf1, 0xfc, 0xf3, 0x3e, 0x3f, 0x3f, 0x3f, 0xf5,
    0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
    0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
    0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
    0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
    0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
    0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
    0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
    0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef
};

vhandle_t	V_GLOBAL
vglStreamLoadBGIFontEx(vhandle_t in,vboolean load_as_koi8)
{
	u32_t initial_pos;
    u32_t glyphs_base;
    u32_t *offsets;
    u32_t i, numchars, firstchar;
    u32_t *recode;
    s8_t top,baseline,bottom;
    u8_t *widths;
    vec_t ascent, descent, height;
    vec_t scale;
    vhandle_t out;
    stroke_font_t *fnt;
    vboolean success;
#ifdef	CHECKED_BUILD
	static char *proc_name = "vglStreamLoadBGIFontEx():";
#endif	/* CHECKED_BUILD */
	initial_pos=fileTell(in);
    if(fileGetLittleLong(in) != 0x08084B50)
    {
#ifdef  CHECKED_BUILD
        conlPrintf(LOG_ERROR,"%s isn't BGI font, magic wrong.\n",
            proc_name);
#endif  /* CHECKED_BUILD */
		fileSeek(in,initial_pos,SEEK_SET);
		return LIBV_INVALID_HANDLE;
    }
    fileSeek(in,initial_pos + 0x81,SEEK_SET);
    /* Read header */
    numchars=fileGetLittleShort(in);
    fileGetc(in);
    firstchar=(unsigned char)fileGetc(in);
    fileSeek(in,3,SEEK_CUR);
    fileRead(in,&top,1);
    fileRead(in,&baseline,1);
    fileRead(in,&bottom,1);
    ascent=(vec_t)top-(vec_t)baseline;
    descent=(vec_t)bottom-(vec_t)baseline;
    height=(vec_t)ascent-(vec_t)descent;
    scale=(vec_t)1 / height;
    height=1;
    ascent*=scale;
    descent*=scale;
    /* Read offset table */
    fileSeek(in,initial_pos + 0x90,SEEK_SET);
    offsets=(u32_t*)memAlloc(numchars*sizeof(u32_t));
    for(i=0;i<numchars;i++)
        offsets[i]=fileGetLittleShort(in);
    /* Read width table */
    widths=(u8_t*)memAlloc(numchars);
    fileRead(in,widths,numchars);
    /* Read strokes */
    success=true;
    out=meshCreateStrokesEx(0,0,0); /* no attribs */
    glyphs_base=fileTell(in);
    for(i=0;i<numchars;i++)
    {
        fileSeek(in,glyphs_base+offsets[i],SEEK_SET);
        success = success && bgi_read_glyph(in,out,scale);
    }
    memFree(offsets);
    if(success)
        fnt = bgi_convert_strokes_to_font(height,ascent,descent,256,out);
    else
        fnt = NULL;
    if(fnt)
    {
        /* Setup width table */
        for(i=0;i<numchars;i++)
            fnt->widths[i]=scale * (vec_t)widths[i];
        /* Setup code table */
        if(load_as_koi8)
            recode=bgi_koi2alt;
        else
            recode=bgi_win2alt;
        memZero(fnt->code_tab,256*sizeof(u32_t));
        for(i=0;i<256;i++)
        {
            if(i<firstchar)
                fnt->code_tab[i]=0;
            else
                fnt->code_tab[i]=recode[i]-firstchar;
        }
    }
    else
    {
#ifdef  CHECKED_BUILD
        conlPrintf(LOG_ERROR,
            "%s unable to read all glyphs.\n",proc_name);
#endif  /* CHECKED_BUILD */
        /* Delete strokes object */
        meshDeleteStrokes(out);
		fileSeek(in,initial_pos,SEEK_SET);
    }
    memFree(widths);
	return (vhandle_t)fnt;
}

vhandle_t	V_GLOBAL
vglLoadBGIFontEx(char *file_name,vboolean load_as_koi8)
{
	vhandle_t in, out;
    in = fileOpenRead(file_name);
    if(in == LIBV_INVALID_HANDLE)
        return LIBV_INVALID_HANDLE;
	out = vglStreamLoadBGIFontEx(in,load_as_koi8);
	fileClose(in);
	return out;
}

vhandle_t	V_GLOBAL
vglLoadBGIFont(char *file_name)
{
	return vglLoadBGIFontEx(file_name,false);
}

#ifdef	CONFIG_GL_PRECOMPILED_FONT

#ifndef	CONFIG_COMPRESSION_ZLIB
#error	CONFIG_GL_PRECOMPILED_FONT depends on CONFIG_COMPRESSION_ZLIB
#endif	/* !CONFIG_COMPRESSION_ZLIB */

#if 0 /* FIXME: update v_config.h */

#ifndef	CONFIG_GL_BGI_FONT
#error	CONFIG_GL_PRECOMPILED_FONT depends on CONFIG_GL_BGI_FONT.
#endif	/* !CONFIG_GZ_IO */

#endif /* 0 */

/*
 * The array below is a hex dump of gzip'ed BGI font 'cyrdrft.bgi'
 */
unsigned char cyrdrft_bgi_gz[] =
{
  0x1F, 0x8B, 0x08, 0x08, 0x32, 0xA4, 0xC5, 0x25, 0x00, 0x03, 0x63, 0x79,
  0x72, 0x64, 0x72, 0x66, 0x74, 0x2E, 0x63, 0x68, 0x72, 0x00, 0xB5, 0x57,
  0x5B, 0x88, 0x24, 0xD5, 0x19, 0xAE, 0xE9, 0x99, 0xE9, 0xAE, 0xE9, 0xEE,
  0xA9, 0xD3, 0x97, 0xA9, 0x71, 0xB2, 0x01, 0x59, 0x24, 0x0F, 0xC1, 0xC8,
  0x62, 0x96, 0x10, 0xD6, 0x20, 0x41, 0x16, 0x59, 0x24, 0xCA, 0x22, 0x22,
  0x22, 0x4B, 0x90, 0x64, 0x11, 0x11, 0x11, 0x23, 0x62, 0x44, 0xC4, 0xC4,
  0xEA, 0x3A, 0xF7, 0x5B, 0x9D, 0xAA, 0xEA, 0xAA, 0xD9, 0xCD, 0x2E, 0xF8,
  0x10, 0x42, 0x08, 0x22, 0x22, 0xE2, 0x93, 0x88, 0x48, 0x1E, 0x42, 0x10,
  0x11, 0x1F, 0x45, 0x42, 0x08, 0x12, 0xF2, 0x10, 0x7C, 0xF0, 0x41, 0x02,
  0x33, 0xBB, 0x33, 0x93, 0xBF, 0xBA, 0xAA, 0x7B, 0x3A, 0x1B, 0x4D, 0xF6,
  0xC5, 0x9E, 0xAE, 0x9E, 0xEE, 0x73, 0xFE, 0xF3, 0xDF, 0x2F, 0xDF, 0x79,
  0xF0, 0x7E, 0xDF, 0x3F, 0xFB, 0xCB, 0x27, 0x7F, 0xFE, 0xDC, 0x2F, 0x5E,
  0x38, 0x79, 0xFF, 0xA9, 0xB3, 0xA7, 0x4E, 0x7E, 0xFF, 0xAE, 0xBB, 0xEE,
  0xDC, 0xEC, 0x7A, 0x27, 0x22, 0xEF, 0xDE, 0x87, 0xCE, 0x3D, 0xFC, 0xD6,
  0x78, 0xC5, 0x5B, 0xF1, 0xBE, 0xC1, 0xD7, 0xF7, 0x3E, 0xF6, 0xBC, 0x93,
  0x51, 0xCB, 0x0B, 0xBD, 0xBD, 0x66, 0xA5, 0xED, 0x9D, 0xF0, 0xEE, 0xF6,
  0x1E, 0xF2, 0x22, 0xEF, 0x4D, 0xEF, 0x33, 0xEF, 0x5F, 0x5E, 0x7F, 0xE5,
  0xB6, 0x95, 0x7B, 0x56, 0xCE, 0xAF, 0x5C, 0x5C, 0x79, 0x7A, 0xE5, 0xE5,
  0x15, 0xB6, 0xF2, 0xC6, 0xCA, 0x3B, 0x2B, 0x7F, 0x59, 0xF1, 0x5B, 0xDB,
  0xAD, 0x33, 0xAD, 0x8B, 0xAD, 0x67, 0x5A, 0xBF, 0x6F, 0x7D, 0xDC, 0xFA,
  0x67, 0xCB, 0x5F, 0xDD, 0x5A, 0xBD, 0x6D, 0xF5, 0xD4, 0xEA, 0xA3, 0xAB,
  0x7F, 0x5A, 0xFD, 0x64, 0xD5, 0x5F, 0x3B, 0xBD, 0xF6, 0xF0, 0xDA, 0x53,
  0x6B, 0x2F, 0xAF, 0xBD, 0xBE, 0xF6, 0xDE, 0xDA, 0x9F, 0xD7, 0x3E, 0x5B,
  0xFB, 0x72, 0xAD, 0xB5, 0xBE, 0xB3, 0x7E, 0xFB, 0xFA, 0x85, 0xF5, 0x17,
  0xD7, 0x5F, 0x5F, 0xFF, 0x68, 0xFD, 0xFA, 0x7A, 0xB7, 0xFD, 0x9D, 0xF6,
  0xE9, 0xF6, 0x7D, 0xED, 0x47, 0xDA, 0x4F, 0xB4, 0x5F, 0x6C, 0x27, 0xED,
  0xCB, 0xED, 0x37, 0xDA, 0xEF, 0xB7, 0x3F, 0x69, 0x7F, 0xDE, 0x1E, 0x75,
  0xCE, 0x74, 0x2E, 0x74, 0xA2, 0xCE, 0x1F, 0x3A, 0xEF, 0x76, 0x3E, 0xEF,
  0xAC, 0xF9, 0x3B, 0xFE, 0x69, 0xFF, 0x5E, 0xFF, 0xBC, 0xFF, 0xBC, 0x6F,
  0xFC, 0xB7, 0xFD, 0x4F, 0x7D, 0x6F, 0x63, 0x6B, 0xE3, 0xEE, 0x8D, 0x07,
  0x37, 0x9E, 0xDC, 0x78, 0x61, 0xC3, 0x6C, 0x5C, 0xD9, 0x78, 0x73, 0xE3,
  0x8F, 0x1B, 0xA3, 0xEE, 0xC9, 0xEE, 0xB3, 0xDD, 0xDF, 0x76, 0x7F, 0xD7,
  0x7D, 0xBB, 0xFB, 0x41, 0xF7, 0x8B, 0xEE, 0x61, 0xF7, 0x44, 0xEF, 0x8E,
  0xDE, 0xB9, 0xDE, 0x13, 0xBD, 0xE7, 0x7B, 0xAC, 0x77, 0xA5, 0xF7, 0x5A,
  0xEF, 0x9D, 0xDE, 0x87, 0xBD, 0x2F, 0x7B, 0xAD, 0xFE, 0x89, 0xFE, 0x3D,
  0xFD, 0xF3, 0xFD, 0x8B, 0x7D, 0xD5, 0xBF, 0xDC, 0x7F, 0xAD, 0xFF, 0x6E,
  0xFF, 0xA3, 0xFE, 0xDF, 0xFB, 0xFF, 0xE8, 0x1F, 0xF6, 0x47, 0x9B, 0x77,
  0x6C, 0x3E, 0xB0, 0xF9, 0xF8, 0x66, 0xBA, 0xF9, 0xCE, 0xE6, 0xA7, 0x9B,
  0xD7, 0x37, 0x6F, 0x0B, 0xCE, 0x06, 0x17, 0x82, 0x97, 0x82, 0x34, 0x78,
  0x2D, 0x78, 0x2F, 0xF8, 0x28, 0xF8, 0x6B, 0xF0, 0x45, 0x30, 0x40, 0xDF,
  0x45, 0xB7, 0xA3, 0x3B, 0xD0, 0x29, 0x74, 0x27, 0x3A, 0x8D, 0x7E, 0x80,
  0x7E, 0x88, 0xCE, 0xA0, 0x1F, 0xA1, 0xBB, 0xD1, 0x8F, 0xD1, 0x3D, 0xE8,
  0x2C, 0xBA, 0x17, 0x9D, 0x43, 0xF7, 0xA1, 0x9F, 0xA0, 0x07, 0xD0, 0x79,
  0xF4, 0x20, 0x7A, 0x08, 0x3D, 0x8C, 0x1E, 0x41, 0x8F, 0xA2, 0x0B, 0xE8,
  0xA7, 0xE8, 0x31, 0xF4, 0x33, 0x74, 0x11, 0x3D, 0x8E, 0x9E, 0x40, 0x4F,
  0xA2, 0xA7, 0xD0, 0xD3, 0xE8, 0x19, 0xF4, 0x2C, 0x7A, 0x0E, 0x3D, 0x8F,
  0x5E, 0x40, 0x2F, 0xA2, 0x97, 0xD0, 0xCB, 0xE8, 0xD7, 0x28, 0x42, 0x31,
  0x22, 0x88, 0x21, 0x81, 0x5E, 0x47, 0x1F, 0xA0, 0x2F, 0xD0, 0x60, 0x70,
  0x66, 0x70, 0x6E, 0xF0, 0xD8, 0xE0, 0xD9, 0x41, 0x3A, 0x78, 0x6B, 0xF0,
  0xF6, 0xE0, 0xC3, 0xC1, 0xDF, 0x06, 0xFB, 0x83, 0x13, 0xC3, 0xC0, 0xEF,
  0x0E, 0xB6, 0x6E, 0xBD, 0x65, 0xA3, 0xDB, 0x45, 0x23, 0x7F, 0xE4, 0x07,
  0xE3, 0xFE, 0x68, 0x34, 0x9E, 0xBD, 0xFD, 0xEE, 0x70, 0x34, 0x1C, 0x7F,
  0x7B, 0x14, 0x86, 0xE1, 0x78, 0x14, 0x6E, 0xFB, 0x83, 0x70, 0xB8, 0xB3,
  0xBD, 0x1D, 0x6E, 0x87, 0x5B, 0x83, 0xED, 0xD1, 0xCE, 0xD6, 0x68, 0xAB,
  0x1B, 0x74, 0xB7, 0x87, 0xDD, 0xF1, 0x18, 0x88, 0xFB, 0xE3, 0x31, 0x90,
  0xFB, 0xB7, 0x8E, 0xE1, 0xB5, 0x39, 0xEC, 0x8F, 0x07, 0xDB, 0xC3, 0x01,
  0xB0, 0x0E, 0x76, 0x26, 0xE1, 0x70, 0x3C, 0xDC, 0x1E, 0x85, 0xC3, 0x21,
  0xFC, 0xDF, 0x19, 0xC3, 0xAF, 0x2D, 0x60, 0xBA, 0xB5, 0x35, 0x0C, 0x6F,
  0x51, 0x5B, 0xC3, 0x21, 0x90, 0x21, 0x78, 0x0D, 0x86, 0x68, 0x38, 0x44,
  0x68, 0x08, 0x5A, 0x4E, 0x26, 0x38, 0xA2, 0x13, 0x16, 0x33, 0x4C, 0x09,
  0xA1, 0x9C, 0x0A, 0xA2, 0xA8, 0xF2, 0x14, 0xE4, 0x51, 0xB4, 0x1E, 0x53,
  0xCC, 0x89, 0x60, 0x52, 0x48, 0x25, 0x34, 0xD3, 0x44, 0xC5, 0x72, 0x82,
  0x06, 0xB7, 0x20, 0x90, 0x85, 0xC6, 0x63, 0x3D, 0xE3, 0x90, 0x78, 0x09,
  0xD0, 0xD2, 0x10, 0x8E, 0x7A, 0x34, 0x26, 0x31, 0x89, 0x68, 0x24, 0x3C,
  0x01, 0x6B, 0x24, 0xC4, 0x25, 0x9E, 0x92, 0x29, 0x29, 0xF1, 0x18, 0x3B,
  0x1E, 0xB2, 0x92, 0x4D, 0xF9, 0x94, 0x97, 0x7C, 0xCC, 0x5D, 0x2D, 0x41,
  0x84, 0x04, 0x68, 0x4D, 0x69, 0x03, 0x9C, 0x60, 0xDF, 0x8A, 0xD4, 0x4B,
  0x61, 0x35, 0x1B, 0xA6, 0x85, 0x2B, 0x69, 0x89, 0x0B, 0x9C, 0x61, 0x87,
  0x0D, 0xD1, 0x4E, 0x65, 0x32, 0xE3, 0x19, 0xCE, 0x26, 0x0E, 0xB8, 0xE3,
  0x09, 0xC6, 0x6A, 0x47, 0x1D, 0x15, 0x5E, 0x01, 0xD4, 0x7A, 0xAC, 0x4A,
  0x90, 0x30, 0xC5, 0x8A, 0x08, 0xD0, 0x52, 0xE9, 0x69, 0x1A, 0xDA, 0xE8,
  0x8A, 0x77, 0x35, 0xBE, 0xAA, 0xAF, 0xD8, 0xC2, 0xE6, 0x3A, 0x8F, 0xF3,
  0xB8, 0x88, 0xAE, 0x44, 0xAF, 0x7A, 0xAF, 0x56, 0xDC, 0x37, 0x32, 0x92,
  0x02, 0x1F, 0x16, 0x91, 0x09, 0x26, 0x58, 0x12, 0xC3, 0x4C, 0x8A, 0xD2,
  0xDC, 0x15, 0xAE, 0xB0, 0x25, 0x2F, 0x69, 0x41, 0x0A, 0x92, 0x13, 0xC7,
  0x4C, 0x11, 0x5D, 0xF2, 0x2E, 0xCD, 0x6C, 0x23, 0x25, 0x99, 0xD2, 0x29,
  0xD8, 0x98, 0x12, 0x27, 0x3D, 0x09, 0x6B, 0x3C, 0x24, 0x05, 0xCE, 0x41,
  0x6B, 0x8E, 0x31, 0x89, 0x78, 0x54, 0xDB, 0x83, 0x43, 0x56, 0xF0, 0x9C,
  0x27, 0x9C, 0x73, 0xCC, 0x22, 0x3C, 0x5F, 0x45, 0xCC, 0x49, 0x64, 0x1C,
  0x0B, 0x79, 0x26, 0x02, 0xA9, 0x64, 0x28, 0x32, 0x1E, 0x30, 0xE5, 0x3C,
  0x57, 0x79, 0x69, 0x23, 0x91, 0x72, 0x28, 0x27, 0xB9, 0x97, 0x57, 0xB2,
  0x3C, 0x12, 0x91, 0x98, 0xC6, 0xF4, 0x90, 0x5C, 0xA7, 0x5E, 0xE3, 0x47,
  0xA0, 0xB8, 0x61, 0xB7, 0xF1, 0xB0, 0x0E, 0xC9, 0x51, 0xED, 0x7F, 0x0E,
  0x3B, 0x18, 0x83, 0x36, 0x09, 0xCE, 0x49, 0xC1, 0x4B, 0x53, 0x26, 0x85,
  0xCB, 0xD3, 0x24, 0xE5, 0x0E, 0x27, 0x91, 0x01, 0x0D, 0xA7, 0xDE, 0xB4,
  0xD2, 0x65, 0x53, 0x96, 0x32, 0x32, 0x9E, 0xA9, 0xBE, 0x0F, 0xC1, 0xAA,
  0xD2, 0x96, 0x6E, 0xEA, 0x32, 0x97, 0x38, 0x6D, 0x15, 0x04, 0x9F, 0x83,
  0x4F, 0x22, 0x17, 0xD5, 0xF2, 0xF0, 0x10, 0x1F, 0x53, 0x38, 0x67, 0xAC,
  0xE6, 0xDA, 0xF6, 0x9C, 0x72, 0xDC, 0x11, 0x37, 0xB1, 0x75, 0x2C, 0x48,
  0x4D, 0x9B, 0xAE, 0x63, 0x9A, 0x94, 0xC9, 0x5C, 0xCE, 0x2A, 0x9E, 0xD0,
  0xC8, 0x46, 0x6E, 0x02, 0x94, 0x0A, 0x4E, 0x26, 0x34, 0x21, 0x06, 0x2B,
  0x5C, 0xBA, 0xB2, 0xA6, 0x77, 0x43, 0x37, 0xB5, 0x55, 0x94, 0xA7, 0x10,
  0x65, 0xB2, 0x44, 0x2D, 0x9C, 0xB2, 0x86, 0x1A, 0xA2, 0x30, 0x6F, 0xB4,
  0x08, 0xD3, 0x92, 0x35, 0x7C, 0xAD, 0x57, 0xC7, 0x1F, 0x0B, 0xAC, 0xA8,
  0xB6, 0xDA, 0x19, 0xD0, 0x2B, 0xAB, 0x39, 0x91, 0x8A, 0x93, 0x23, 0x96,
  0xD6, 0x3A, 0x0A, 0x87, 0x2B, 0x1D, 0x1B, 0x1E, 0x6B, 0x4B, 0x12, 0xB2,
  0x25, 0xC9, 0x16, 0xF8, 0x48, 0x2B, 0x81, 0xDE, 0xD6, 0x94, 0xA4, 0x4F,
  0x40, 0x57, 0x90, 0x6F, 0x48, 0xAB, 0xF2, 0x74, 0xE5, 0xF1, 0x26, 0x0E,
  0x8B, 0x1D, 0xDA, 0xAA, 0xF3, 0x9C, 0xB6, 0xAA, 0x38, 0xD5, 0x51, 0x4E,
  0x86, 0x58, 0x26, 0x93, 0xCC, 0xCB, 0x66, 0x94, 0x89, 0x21, 0xED, 0x84,
  0xCD, 0xBD, 0x98, 0x48, 0x3C, 0xDF, 0x09, 0x48, 0x46, 0xA1, 0x06, 0x92,
  0x32, 0x9D, 0xA6, 0x19, 0xC4, 0xC7, 0x24, 0x4A, 0x2B, 0x25, 0x14, 0x07,
  0x2E, 0x31, 0xFC, 0x45, 0xAA, 0xB1, 0x34, 0xDF, 0x04, 0x97, 0xA7, 0x26,
  0x55, 0x4E, 0x26, 0x42, 0x0B, 0x21, 0x99, 0x06, 0xE1, 0x29, 0xCB, 0x84,
  0x19, 0xA8, 0x44, 0x6A, 0x29, 0x14, 0xAC, 0xE4, 0x83, 0x4C, 0x64, 0x0C,
  0x82, 0x44, 0x2F, 0xF1, 0xCB, 0xEA, 0xB2, 0xB9, 0xE4, 0x76, 0xB3, 0x62,
  0x9A, 0x43, 0xFD, 0x98, 0x52, 0x16, 0x7C, 0x4A, 0x33, 0x02, 0x15, 0x04,
  0x59, 0xCE, 0x28, 0xE1, 0x50, 0xBD, 0xA6, 0x8A, 0xED, 0xA4, 0x88, 0xC1,
  0xF8, 0x41, 0x2E, 0x72, 0x38, 0x79, 0xD5, 0xBB, 0x0A, 0xD2, 0x64, 0x38,
  0x89, 0x64, 0x98, 0x45, 0xA4, 0x93, 0x34, 0x5E, 0x27, 0x50, 0x9B, 0x24,
  0xB4, 0x65, 0x5A, 0x64, 0xD3, 0x3C, 0xCB, 0x5D, 0x66, 0x53, 0x88, 0x3F,
  0xE9, 0x59, 0x9D, 0x42, 0x35, 0xE6, 0x3C, 0x27, 0x59, 0x9C, 0x82, 0x7F,
  0x49, 0x54, 0x7A, 0x65, 0xA5, 0xEF, 0x00, 0x88, 0xA0, 0x82, 0x64, 0xC9,
  0x0B, 0x9A, 0x93, 0x14, 0x3C, 0x2B, 0x08, 0x85, 0x06, 0x32, 0x91, 0x33,
  0xBF, 0x67, 0x38, 0xA7, 0x35, 0x65, 0xCD, 0x59, 0x43, 0x86, 0xA6, 0x79,
  0x96, 0xE6, 0x36, 0x17, 0x19, 0x4D, 0x71, 0x32, 0xD1, 0x0B, 0x5E, 0x35,
  0x45, 0x56, 0x92, 0x9E, 0xD1, 0xC4, 0xCB, 0x1A, 0x8F, 0x2C, 0xAF, 0xD6,
  0x3A, 0xDE, 0x8C, 0xCC, 0x5C, 0x58, 0x3F, 0x17, 0xC7, 0x7C, 0xF3, 0x30,
  0x8F, 0x48, 0x2F, 0xD7, 0xBB, 0xDE, 0x6E, 0xB3, 0x52, 0x47, 0xD7, 0x84,
  0x86, 0xEA, 0x58, 0x4D, 0x44, 0xDD, 0x19, 0x62, 0x28, 0x33, 0x5E, 0xF7,
  0xA3, 0xFA, 0x14, 0xE1, 0xB2, 0x9F, 0x2F, 0xEB, 0xE7, 0xA5, 0x51, 0x13,
  0xD3, 0x99, 0x5E, 0x26, 0x2A, 0x66, 0x4F, 0x11, 0x5D, 0xF6, 0x2E, 0x2F,
  0x56, 0xF3, 0x99, 0xBC, 0x5A, 0x96, 0x0C, 0xBF, 0x5E, 0xCB, 0xA9, 0x98,
  0xDA, 0x3C, 0x9D, 0x5B, 0x73, 0xAC, 0xDB, 0x52, 0x04, 0x92, 0xCC, 0xA4,
  0x50, 0xA3, 0x44, 0xD5, 0x3A, 0xDC, 0x3C, 0x37, 0xB3, 0x96, 0x1F, 0x7E,
  0x25, 0xC7, 0x79, 0x4C, 0xB5, 0xEE, 0xCD, 0x2D, 0xCB, 0x46, 0xAE, 0x30,
  0xA5, 0x80, 0x2E, 0x08, 0xDD, 0x2D, 0x85, 0x7C, 0xB7, 0xDC, 0x58, 0x48,
  0xC1, 0x14, 0xB2, 0x0C, 0x8A, 0x0B, 0xF2, 0x47, 0xC0, 0xB4, 0xC0, 0xB8,
  0xEE, 0xBD, 0x22, 0x14, 0xD1, 0x24, 0x74, 0xE5, 0xC2, 0x4F, 0xAC, 0xD2,
  0x43, 0x41, 0xBF, 0x99, 0xE9, 0xC1, 0xF2, 0xC6, 0x92, 0x49, 0x28, 0xA3,
  0x0C, 0x9E, 0x3A, 0x6A, 0x71, 0xC8, 0x23, 0x33, 0x7B, 0xB2, 0x68, 0x17,
  0x9E, 0xDA, 0x5F, 0x18, 0xBE, 0x65, 0x21, 0x8E, 0x6A, 0xCE, 0x70, 0x42,
  0xCF, 0xCE, 0x34, 0x91, 0xAE, 0x76, 0x80, 0xA2, 0xC4, 0x90, 0x0D, 0x35,
  0x05, 0xCC, 0x94, 0x23, 0x32, 0x26, 0x47, 0x78, 0x2C, 0xA6, 0xF8, 0x15,
  0x71, 0xD4, 0xCC, 0xAC, 0x30, 0x39, 0xA8, 0x3B, 0x22, 0x1B, 0xB3, 0x23,
  0x98, 0x34, 0x47, 0xF1, 0x98, 0x4F, 0xE3, 0x57, 0x78, 0xB3, 0x8F, 0x3B,
  0xDA, 0x4C, 0x39, 0x7C, 0xEA, 0x29, 0xAF, 0x75, 0xC3, 0x9B, 0xD8, 0x51,
  0x37, 0xEB, 0x0D, 0xB1, 0x13, 0xD0, 0x03, 0x55, 0x1D, 0x55, 0x16, 0xD2,
  0x6A, 0x0E, 0xA4, 0xD4, 0xB1, 0x94, 0x66, 0xF5, 0x69, 0x17, 0xB8, 0xC8,
  0xF5, 0xAC, 0xD5, 0x89, 0x48, 0x98, 0x25, 0x1A, 0xBA, 0x10, 0x23, 0x98,
  0x41, 0xCE, 0xE8, 0xC8, 0x4E, 0x1C, 0x5E, 0xCA, 0xD3, 0x1E, 0xB3, 0x22,
  0xD1, 0x89, 0xB5, 0x4E, 0xA7, 0x22, 0x65, 0x0E, 0x5B, 0xC8, 0x6F, 0xC8,
  0xAD, 0x09, 0x69, 0xA8, 0xFE, 0x17, 0x9F, 0xA6, 0x43, 0x86, 0x37, 0x23,
  0x0D, 0xFB, 0x55, 0xC7, 0x4C, 0xCC, 0xFF, 0xE3, 0xA6, 0x42, 0x51, 0xB2,
  0x82, 0x66, 0x34, 0x8A, 0x03, 0x99, 0xD4, 0x13, 0x00, 0xEC, 0x39, 0x4C,
  0xAE, 0xD9, 0x7D, 0xBD, 0x27, 0xF6, 0xD8, 0xFE, 0x4D, 0x5A, 0xD6, 0xE5,
  0x16, 0xCE, 0x27, 0x89, 0x85, 0x9E, 0x39, 0xEF, 0xF6, 0x30, 0x0D, 0xA1,
  0xFF, 0xEE, 0xE2, 0x12, 0xFA, 0x5B, 0x53, 0x4F, 0xB4, 0x9A, 0x85, 0x25,
  0xDB, 0xA5, 0x25, 0x0B, 0xD8, 0x01, 0xDD, 0xC7, 0x7B, 0x93, 0xBD, 0xDA,
  0x8B, 0x15, 0x97, 0x24, 0x20, 0x44, 0xF8, 0x6E, 0xA9, 0x82, 0x9A, 0x1E,
  0x1B, 0xDC, 0x28, 0xC1, 0x75, 0x73, 0x5B, 0x24, 0x97, 0x92, 0xDF, 0xD8,
  0x2B, 0x6A, 0x3E, 0xC3, 0xFF, 0x9B, 0xAA, 0xD6, 0x43, 0x04, 0x5F, 0xA5,
  0x7B, 0xCA, 0x52, 0x01, 0xD3, 0x6D, 0x66, 0x5B, 0x63, 0x47, 0x40, 0xF6,
  0x6E, 0x22, 0x42, 0x81, 0xDB, 0xBB, 0x09, 0x9F, 0x54, 0xBA, 0xF8, 0x54,
  0xD7, 0xDA, 0x58, 0x98, 0x50, 0x30, 0x0B, 0xE0, 0x94, 0x4A, 0x78, 0x42,
  0x2C, 0xD6, 0x44, 0x32, 0x01, 0x9B, 0x2C, 0x21, 0x09, 0x48, 0x50, 0x30,
  0x89, 0xA1, 0xBB, 0xE0, 0xDA, 0x6E, 0x40, 0x14, 0xA4, 0xE2, 0xA7, 0x96,
  0x62, 0x02, 0xFC, 0x08, 0x9D, 0x00, 0xA2, 0x88, 0x4C, 0x35, 0xAB, 0x82,
  0xB9, 0x6D, 0x71, 0x20, 0xC0, 0x6B, 0x22, 0xAA, 0x6B, 0x0D, 0x07, 0x3C,
  0xD2, 0xB3, 0x07, 0xF6, 0xE1, 0x69, 0x72, 0x39, 0x48, 0x80, 0x06, 0x37,
  0x5E, 0x9D, 0x9F, 0x60, 0xD7, 0xC9, 0x7E, 0x0C, 0xDE, 0xAF, 0x4F, 0x56,
  0xFB, 0x40, 0x97, 0x60, 0x2F, 0x69, 0xE8, 0xE4, 0xB7, 0xF8, 0x65, 0x76,
  0x09, 0x46, 0xC8, 0x94, 0x65, 0x3C, 0x15, 0x89, 0x30, 0x4C, 0xF1, 0x1D,
  0xB6, 0x5B, 0x61, 0x19, 0x91, 0x49, 0x27, 0xAD, 0x00, 0x2B, 0x04, 0x97,
  0x54, 0x62, 0x01, 0x9A, 0xB1, 0x43, 0x76, 0x9D, 0xEF, 0x33, 0x5F, 0x30,
  0x80, 0x0B, 0x31, 0x9B, 0xD0, 0x23, 0x7A, 0xC0, 0xAE, 0xF1, 0x7D, 0xB9,
  0x57, 0x57, 0x1E, 0x44, 0xD3, 0x90, 0x8D, 0x45, 0x1E, 0x00, 0x77, 0x71,
  0x09, 0x50, 0xC7, 0x14, 0x50, 0x4F, 0xCA, 0x12, 0x66, 0x60, 0xD8, 0xED,
  0x88, 0x5D, 0x01, 0xDC, 0x59, 0x06, 0x95, 0x67, 0x99, 0x56, 0x92, 0x71,
  0x0A, 0x3D, 0x8C, 0x01, 0x77, 0x71, 0x28, 0x80, 0xBB, 0xF0, 0x19, 0x63,
  0xC0, 0x5D, 0x4C, 0xE4, 0x91, 0x3C, 0x10, 0xC0, 0x9D, 0x36, 0xDC, 0x71,
  0x1B, 0xFA, 0x9D, 0x06, 0x2E, 0x46, 0x41, 0x60, 0x52, 0x80, 0x5B, 0x85,
  0x28, 0x15, 0xF6, 0x09, 0x8C, 0x43, 0x80, 0x7F, 0xE0, 0xF6, 0x94, 0xE5,
  0x55, 0xEA, 0xA9, 0xD2, 0xD4, 0x9D, 0x05, 0xCE, 0x78, 0xAA, 0xCC, 0xA3,
  0x74, 0x9D, 0x36, 0x13, 0xC8, 0x56, 0xC8, 0xB4, 0xAE, 0x78, 0xC0, 0x1A,
  0xD8, 0xD4, 0x7E, 0xC0, 0x1E, 0x3E, 0x46, 0x43, 0xD0, 0x17, 0x71, 0x85,
  0x23, 0x64, 0xCA, 0xD3, 0x19, 0xD6, 0xC1, 0xF3, 0x5C, 0xF7, 0x2A, 0x0C,
  0x93, 0x4C, 0x9B, 0x33, 0xBF, 0xC2, 0x15, 0x0E, 0xC8, 0x79, 0x99, 0x95,
  0x30, 0xA7, 0xA2, 0xE9, 0x41, 0x06, 0x08, 0xA5, 0x8E, 0x86, 0xF3, 0xC0,
  0xD3, 0x15, 0x6D, 0x0F, 0xEB, 0xF9, 0x4C, 0xD5, 0x5A, 0x97, 0x79, 0xF5,
  0x19, 0x81, 0x4E, 0x66, 0xEA, 0x19, 0x53, 0x6B, 0x84, 0x47, 0x14, 0x26,
  0xB6, 0xCB, 0x9D, 0xD5, 0x1A, 0xEA, 0x39, 0x36, 0x15, 0xB6, 0x8A, 0x65,
  0x4F, 0xEB, 0x46, 0x0E, 0xF4, 0x3F, 0x57, 0xCE, 0xEB, 0x65, 0xFE, 0x8B,
  0x6F, 0xAB, 0xDD, 0x63, 0xDD, 0x5D, 0xC8, 0x40, 0x67, 0xD6, 0x4B, 0x17,
  0x9A, 0xAA, 0x29, 0xF0, 0x3C, 0x3E, 0x05, 0x34, 0x26, 0x2A, 0xCB, 0x72,
  0xDE, 0x71, 0xE1, 0x37, 0xEE, 0xA5, 0x3A, 0x0D, 0xE7, 0x27, 0x6E, 0x44,
  0x91, 0xB3, 0x49, 0x91, 0xE7, 0x49, 0x0E, 0xD8, 0x3B, 0x05, 0x7F, 0xF1,
  0x66, 0x4A, 0xD4, 0x3E, 0x58, 0xE6, 0x0B, 0x7E, 0x2B, 0x00, 0xCF, 0x18,
  0xF0, 0xA7, 0xC4, 0xB2, 0xE6, 0x96, 0xCD, 0x50, 0x75, 0x85, 0xA9, 0x17,
  0x58, 0x39, 0xAE, 0xB0, 0x32, 0xF8, 0x3D, 0x4A, 0x71, 0x46, 0x9B, 0x3E,
  0x0E, 0xF3, 0x43, 0x2E, 0x66, 0x03, 0x58, 0x66, 0x24, 0x03, 0x54, 0x57,
  0xCA, 0xF6, 0x3C, 0x8F, 0x71, 0x07, 0x46, 0x5A, 0x21, 0x0A, 0x95, 0x6B,
  0xC0, 0xE8, 0x59, 0x9E, 0x71, 0x47, 0x35, 0x55, 0x4C, 0x54, 0x29, 0xC3,
  0x55, 0xA8, 0x16, 0xF6, 0xC2, 0x7C, 0x08, 0x17, 0xF3, 0x01, 0x7C, 0x04,
  0x13, 0xE5, 0xA0, 0xEA, 0xA2, 0x8B, 0x15, 0x45, 0x39, 0x60, 0xD7, 0x70,
  0xD9, 0x8F, 0x10, 0xB3, 0x52, 0x87, 0x7A, 0x6E, 0xD9, 0xF2, 0x4A, 0x54,
  0x1E, 0xD4, 0x37, 0x80, 0x7A, 0xDD, 0xCC, 0xF2, 0xC5, 0x54, 0xF9, 0xF2,
  0x1F, 0x52, 0x96, 0xD6, 0x6B, 0xAE, 0xF3, 0x58, 0x36, 0x51, 0x64, 0x3D,
  0xA7, 0x8F, 0x3D, 0x85, 0xBB, 0x4A, 0xA9, 0xDC, 0x94, 0x79, 0x59, 0xE4,
  0x45, 0x9C, 0xC3, 0x69, 0x40, 0x7D, 0x6A, 0x9E, 0x31, 0xB3, 0x7B, 0x4F,
  0x8E, 0x35, 0x85, 0xB2, 0x83, 0xF8, 0xC9, 0xE6, 0x5C, 0x85, 0x3C, 0x2B,
  0x18, 0x9D, 0x58, 0x63, 0x23, 0xDB, 0x36, 0x5C, 0x89, 0x19, 0x56, 0x67,
  0x78, 0x86, 0x43, 0xAB, 0x89, 0x6B, 0xE3, 0x66, 0x0A, 0x07, 0x6C, 0x4E,
  0x89, 0x0D, 0xF4, 0x9B, 0x19, 0x9A, 0xC1, 0x70, 0x8B, 0x4A, 0x58, 0x2A,
  0x33, 0x9D, 0x1B, 0xC0, 0xBE, 0xF5, 0xCD, 0xA3, 0xD2, 0x86, 0x95, 0x62,
  0x0A, 0xDD, 0x0D, 0x74, 0x67, 0x7D, 0x05, 0x93, 0xDE, 0x02, 0x02, 0x85,
  0xDC, 0xAA, 0xF7, 0xFB, 0x34, 0xE1, 0x35, 0x2F, 0xAD, 0x01, 0x41, 0x56,
  0x28, 0x9D, 0x41, 0x87, 0xD4, 0x13, 0x1B, 0xD7, 0x14, 0xB6, 0x0F, 0xF0,
  0xD5, 0x31, 0x07, 0xAD, 0xCD, 0xC0, 0x3D, 0x08, 0x76, 0xAB, 0xAE, 0x65,
  0xB1, 0xCD, 0x4D, 0xA1, 0x4A, 0x3C, 0x97, 0x03, 0x37, 0x3B, 0xE0, 0x0C,
  0x60, 0x9D, 0x02, 0x3A, 0x87, 0x3B, 0x00, 0x25, 0x80, 0xC5, 0xF5, 0x42,
  0x63, 0x8C, 0x20, 0xE2, 0x70, 0x43, 0x92, 0x91, 0xE8, 0xC0, 0x9D, 0xA9,
  0xB3, 0x88, 0xCE, 0x26, 0x75, 0xDA, 0x59, 0x68, 0xA5, 0x92, 0x5B, 0x6A,
  0x63, 0x3D, 0xF3, 0x26, 0xEF, 0x48, 0xDE, 0xF0, 0x45, 0x95, 0xF5, 0x15,
  0x1F, 0xEB, 0x6C, 0x6B, 0xDE, 0xCD, 0x6E, 0x5C, 0x65, 0x03, 0x99, 0x2E,
  0x79, 0xDF, 0xA7, 0x02, 0xF4, 0xF0, 0xCD, 0xDC, 0x46, 0x8F, 0x1B, 0xE9,
  0x80, 0x72, 0xF1, 0x1B, 0xBB, 0x4A, 0x9B, 0xA5, 0xDC, 0x76, 0xD0, 0xD3,
  0x1D, 0xEE, 0xD8, 0x46, 0x2A, 0xDC, 0xC1, 0x26, 0x98, 0xC2, 0x6D, 0xD3,
  0xCE, 0xB5, 0x03, 0xDD, 0x26, 0x95, 0x6E, 0x8D, 0x35, 0x70, 0x02, 0x77,
  0xAB, 0xDD, 0xEA, 0xDA, 0x03, 0xBB, 0xCD, 0xFA, 0x37, 0xFD, 0x52, 0x68,
  0x16, 0xF3, 0xB8, 0x8A, 0x39, 0x9D, 0x21, 0xD8, 0x2A, 0xE6, 0x95, 0x16,
  0xE0, 0x91, 0xFD, 0x26, 0x5E, 0xDD, 0x0A, 0xCC, 0x2D, 0x47, 0x61, 0x16,
  0x2F, 0x72, 0xA3, 0xE6, 0x70, 0x1D, 0xB2, 0x70, 0x93, 0x8E, 0x34, 0xD0,
  0x27, 0xB3, 0xEB, 0x52, 0xA1, 0xE6, 0xB7, 0x61, 0xE0, 0x76, 0x1C, 0xE9,
  0xD8, 0x22, 0x7B, 0x60, 0xAE, 0xA9, 0x7D, 0x98, 0xEF, 0xD7, 0x1A, 0x0F,
  0x6E, 0x12, 0x28, 0x31, 0x09, 0x72, 0xAC, 0x83, 0x37, 0x74, 0x4E, 0x15,
  0x49, 0xE8, 0xEB, 0xA4, 0xBA, 0xB7, 0x59, 0x88, 0xF2, 0xFE, 0xDC, 0xB3,
  0xE0, 0x55, 0x34, 0xF7, 0xCE, 0x8D, 0x51, 0x4B, 0x0E, 0x17, 0xD1, 0x14,
  0x94, 0x59, 0x06, 0x72, 0xA2, 0xE5, 0xA8, 0xF3, 0x48, 0xC6, 0x50, 0x22,
  0x2D, 0x55, 0x77, 0x71, 0xE7, 0x5A, 0x8B, 0xAE, 0xF7, 0xB5, 0xFB, 0xE9,
  0x61, 0x4D, 0x31, 0xA3, 0xA9, 0xF4, 0x8F, 0x61, 0x14, 0x29, 0xAC, 0x1C,
  0x72, 0xC7, 0xB9, 0x03, 0xEB, 0xA0, 0x43, 0xBD, 0xDE, 0xD8, 0xD3, 0xC4,
  0x79, 0x91, 0x7F, 0xF3, 0x4C, 0x98, 0xF9, 0xAB, 0x03, 0x77, 0x7D, 0x2B,
  0x6B, 0x5B, 0x63, 0x1B, 0xC9, 0x88, 0xC7, 0x9C, 0xD7, 0xDC, 0xAA, 0xBC,
  0x81, 0xDB, 0x57, 0x85, 0xCF, 0xC1, 0x02, 0xC8, 0x34, 0x56, 0x9F, 0xFB,
  0x37, 0xCD, 0xEF, 0x49, 0x44, 0x32, 0x14, 0x00, 0x00
};

void	V_GLOBAL
vglLoadDefaultFontEx(vboolean load_as_koi8)
{
	vhandle_t glctx;
	vhandle_t in, fnt;
	io_section_p compressed;
	glctx = vglGetCurrentWindow();
	if(glctx == LIBV_INVALID_HANDLE)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vglLoadDefaultFontEx(): OpenGL context not initialized.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	/*
	 * Transform array into file, decompressing it on the fly.
	 * Looks great, isn't it ? :)
	 */
	compressed = ioAllocSection(sizeof(io_section_t),cyrdrft_bgi_gz,
		sizeof(cyrdrft_bgi_gz));
	/*
	 * We don't need any cleanups, so no warning messages, please
	 */
	compressed->vtbl.base.destroy = libv_default_destructor;
	in = gzConvertToDecompress(compressed);
	fnt = vglStreamLoadBGIFontEx(in,load_as_koi8);

	fileClose(in);
	vglSelectFont(fnt);
	/*
	 * Reference counter is now 2; decrease it:
	 */
	vglDestroyFont(fnt);
	/*
	 * Font handle will be closed either by next call to
	 * vglMakeFontCurrent(), or by vglDeleteWindow().
	 */
}

void	V_GLOBAL
vglLoadDefaultFont()
{
	vglLoadDefaultFontEx(false);
}

#endif	/* CONFIG_PRECOMPILED_FONT */
